#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#if defined(__OpenBSD__)
# include <sys/dvdio.h>
#elif defined(__linux__)
# include <linux/cdrom.h>
#else
# error "Need the DVD ioctls"
#endif
#include "css_auth.h"

static u_char Challenge[10];
static struct block Key1;
static struct block Key2;
static struct block KeyCheck;
static int varient = -1;

#ifdef DEBUG

/**
 *
 */

static void _CSSPrintChallene (const u_char *chal)
{
	int i;

	for (i=0; i<10; ++i)
		printf (" %02X", chal[9-i] & 0xff);
}


/**
 *
 */

static void _CSSPrintKey (const u_char *key)
{
	int i;

	for (i=0; i<5; ++i)
		printf (" %02X", key[4-i] & 0xff);
}


/**
 *
 */

static void _CSSPrint5 (const u_char *key)
{
	int i;

	for (i=0; i<5; ++i)
		printf (" %02X", key[i] & 0xff);
}
#endif


/**
 *
 */

static int _CSSAuthDrive (const u_char *key)
{
	int i;

	for (i=0; i<5; i++)
		Key1.b[i] = key[4-i];

	for (i=0; i<32; ++i) {
		CryptKey1 (i, Challenge, &KeyCheck);

		if (memcmp (KeyCheck.b, Key1.b, 5)==0) {
			varient = i;
#ifdef DEBUG
			printf ("Drive Authentic - using varient %d\n", i);
#endif
			return 0;
		}
	}

#ifdef DEBUG
	if (varient == -1)
		printf ("Drive would not Authenticate\n");
#endif

	return -1;
}


/**
 *
 */

static int _CSSGetDiskKey (int fd, int agid, char *key, char *key_disc)
{
	dvd_struct s;
	int i;

	s.type = DVD_STRUCT_DISCKEY;
	s.disckey.agid = agid;
	memset (s.disckey.value, 0, 2048);

	if (ioctl(fd, DVD_READ_STRUCT, &s)<0) {
		perror ("Could not read Disc Key\n");
		return -1;
	}

	for (i=0; i<sizeof s.disckey.value; i++)
		s.disckey.value[i] ^= key[4 - (i%5)];

#ifdef DEBUG
	printf ("Received Disc Key:\t");
	for (i=0; i<10; i++) 
		printf ("%02X ", s.disckey.value[i]);
	printf ("\n");
#endif

	memcpy (key_disc, s.disckey.value, 2048);

	return 0;
}


/**
 *
 */
 
static int _get_title_key (int fd, int agid, int lba, char *key, char *key_title)
{
	dvd_authinfo ai;
	int i;

	ai.type = DVD_LU_SEND_TITLE_KEY;

	ai.lstk.agid = agid;
	ai.lstk.lba = lba;

	if (ioctl (fd, DVD_AUTH, &ai)) {
		perror ("GetTitleKey failed\n");
		return -1;
	}

	for (i=0; i<5; ++i)
		ai.lstk.title_key[i] ^= key[4 - (i%5)];

#ifdef DEBUG
	printf ("Received Title Key:\t");
	for (i=0; i<5; ++i)
		printf ("%02X ", ai.lstk.title_key[i]);
	printf ("\n");

	printf (" CPM=%d, CP_SEC=%d, CGMS=%d\n", ai.lstk.cpm, ai.lstk.cp_sec, ai.lstk.cgms);
#endif

	memcpy (key_title, ai.lstk.title_key, 5);

	return 0;
}


/**
 *
 */

int CSSGetASF (int fd)
{
	dvd_authinfo ai;

	ai.type = DVD_LU_SEND_ASF;
	ai.lsasf.agid = 0;
	ai.lsasf.asf = 0;

	if (ioctl(fd, DVD_AUTH, &ai)) {
		perror ("GetASF");
		return -1;
	}

#ifdef DEBUG
	printf("%sAuthenticated\n", (ai.lsasf.asf) ? "" : "not ");
#endif

	return 0;
}


/**
 * Simulation of a non-CSS compliant host (i.e. the authentication fails,
 * but idea is here for a real CSS compliant authentication scheme).
 */

static int _CSSAuthHost (dvd_authinfo *ai)
{
	int i;

	switch (ai->type) {
	/* Host data receive (host changes state) */
	case DVD_LU_SEND_AGID:
#ifdef DEBUG
		printf ("AGID %d\n", ai->lsa.agid);
#endif
		ai->type = DVD_HOST_SEND_CHALLENGE;
		break;

	case DVD_LU_SEND_KEY1:
#ifdef DEBUG
		printf ("LU sent key1: "); _CSSPrintKey (ai->lsk.key); printf("\n");
#endif
		if (_CSSAuthDrive (ai->lsk.key)<0) {
			ai->type = DVD_AUTH_FAILURE;
			return -EINVAL;
		}
		ai->type = DVD_LU_SEND_CHALLENGE;
		break;

	case DVD_LU_SEND_CHALLENGE:
		for (i = 0; i < 10; ++i)
			Challenge[i] = ai->hsc.chal[9-i];
#ifdef DEBUG
		printf ("LU sent challenge: "); _CSSPrintChallene(Challenge); printf("\n");
#endif
		CryptKey2 (varient, Challenge, &Key2);
		ai->type = DVD_HOST_SEND_KEY2;
		break;

	/* Host data send */
	case DVD_HOST_SEND_CHALLENGE:
		for (i = 0; i < 10; ++i)
			ai->hsc.chal[9-i] = Challenge[i];
#ifdef DEBUG
		printf ("Host sending challenge: "); _CSSPrintChallene(Challenge); printf("\n");
#endif
		/* Returning data, let LU change state */
		break;

	case DVD_HOST_SEND_KEY2:
		for (i = 0; i < 5; ++i)
			ai->hsk.key[4-i] = Key2.b[i];
#ifdef DEBUG
		printf ("Host sending key 2: "); _CSSPrintKey (Key2.b); printf("\n");
#endif
		/* Returning data, let LU change state */
		break;

	default:
#ifdef DEBUG
		printf ("Got invalid state %d\n", ai->type);
#endif
		return -EINVAL;
	}

	return 0;
}


/**
 *
 */

int CSSAuthDisk (int fd, char *key)
{
	dvd_authinfo ai;
	dvd_struct dvds;
	int i, rv, tries, agid;

	memset (&ai, 0, sizeof (ai));
	memset (&dvds, 0, sizeof (dvds));

	if (CSSGetASF (fd) < 0)
		return -1;

	/* Init sequence, request AGID */
	for (tries = 1, rv = -1; rv == -1 && tries < 4; ++tries) {
#ifdef DEBUG
		printf ("Request AGID [%d]...\t", tries);
#endif
		ai.type = DVD_LU_SEND_AGID;
		ai.lsa.agid = 0;
		rv = ioctl(fd, DVD_AUTH, &ai);

		if (rv == -1) {
			perror("N/A, invalidating");
			ai.type = DVD_INVALIDATE_AGID;
			ai.lsa.agid = 0;
			ioctl(fd, DVD_AUTH, &ai);
		}
	}
	if (tries == 4) {
#ifdef DEBUG
		printf ("Cannot get AGID\n");
#endif
		return -1;
	}

	for (i = 0; i < 10; ++i)
		Challenge[i] = i;

	/* Send AGID to host */
	if (_CSSAuthHost (&ai) < 0) {
#ifdef DEBUG
		printf ("Send AGID to host failed\n");
#endif
		return -1;
	}
	/* Get challenge from host */
	if (_CSSAuthHost (&ai) < 0) {
#ifdef DEBUG
		printf ("Get challenge from host failed\n");
#endif
		return -1;
	}
	agid = ai.lsa.agid;
	/* Send challenge to LU */
	if (ioctl (fd, DVD_AUTH, &ai) < 0) {
#ifdef DEBUG
		printf ("Send challenge to LU failed\n");
#endif
		return -1;
	}
	/* Get key1 from LU */
	if (ioctl (fd, DVD_AUTH, &ai) < 0) {
#ifdef DEBUG
		printf ("Get key1 from LU failed\n");
#endif
		return -1;
	}
	/* Send key1 to host */
	if (_CSSAuthHost (&ai) < 0) {
#ifdef DEBUG
		printf ("Send key1 to host failed\n");
#endif
		return -1;
	}
	/* Get challenge from LU */
	if (ioctl (fd, DVD_AUTH, &ai) < 0) {
#ifdef DEBUG
		printf ("Get challenge from LU failed\n");
#endif
		return -1;
	}
	/* Send challenge to host */
	if (_CSSAuthHost (&ai) < 0) {
#ifdef DEUG
		printf ("Send challenge to host failed\n");
#endif
		return -1;
	}
	/* Get key2 from host */
	if (_CSSAuthHost (&ai) < 0) {
#ifdef DEBUG
		printf ("Get key2 from host failed\n");
#endif
		return -1;
	}
	/* Send key2 to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
#ifdef DEBUG
		printf ("Send key2 to LU failed (expected)\n");
#endif
		return -1;
	}

#ifdef LIB
	if (ai.type == DVD_AUTH_FAILURE)
		return -1;	
#else
	if (ai.type == DVD_AUTH_ESTABLISHED)
		printf ("DVD is authenticated\n");
	else if (ai.type == DVD_AUTH_FAILURE)
		printf ("DVD authentication failed\n");
#endif
	memcpy (Challenge, Key1.b, 5);
	memcpy (Challenge+5, Key2.b, 5);
	CryptBusKey (varient, Challenge, &KeyCheck);

#ifdef DEBUG
	printf ("Received Session Key:\t");
	for (i= 0; i< 5; i++) {
		printf("%02X ", KeyCheck.b[i]);
	}
	printf ("\n");
#endif

	if (CSSGetASF (fd) < 0)
		return -1;

	_CSSGetDiskKey (fd, agid, KeyCheck.b, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}


/**
 *
 */

int CSSAuthTitle (int fd, char *key, int lba)
{
	dvd_authinfo ai;
	dvd_struct dvds;
	int i, rv, tries, agid;

	memset (&ai, 0, sizeof (ai));
	memset (&dvds, 0, sizeof (dvds));

	if (CSSGetASF (fd))
		return -1;

	/* Init sequence, request AGID */
	for (tries = 1, rv = -1; rv == -1 && tries < 4; ++tries) {
		printf ("Request AGID [%d]...\t", tries);
		ai.type = DVD_LU_SEND_AGID;
		ai.lsa.agid = 0;
		rv = ioctl (fd, DVD_AUTH, &ai);
		if (rv == -1) {
			perror ("N/A, invalidating");
			ai.type = DVD_INVALIDATE_AGID;
			ai.lsa.agid = 0;
			ioctl (fd, DVD_AUTH, &ai);
		}
	}

	if (tries == 4) {
		printf ("Cannot get AGID\n");
		return -1;
	}

	for (i = 0; i < 10; ++i)
		Challenge[i] = i;

	/* Send AGID to host */
	if (_CSSAuthHost(&ai) < 0) {
		printf ("Send AGID to host failed\n");
		return -1;
	}
	/* Get challenge from host */
	if (_CSSAuthHost(&ai) < 0) {
		printf ("Get challenge from host failed\n");
		return -1;
	}
	agid = ai.lsa.agid;
	/* Send challenge to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		printf ("Send challenge to LU failed\n");
		return -1;
	}
	/* Get key1 from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		printf ("Get key1 from LU failed\n");
		return -1;
	}
	/* Send key1 to host */
	if (_CSSAuthHost(&ai) < 0) {
		printf ("Send key1 to host failed\n");
		return -1;
	}
	/* Get challenge from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		printf ("Get challenge from LU failed\n");
		return -1;
	}
	/* Send challenge to host */
	if (_CSSAuthHost (&ai) < 0) {
		printf ("Send challenge to host failed\n");
		return -1;
	}
	/* Get key2 from host */
	if (_CSSAuthHost (&ai) < 0) {
		printf ("Get key2 from host failed\n");
		return -1;
	}
	/* Send key2 to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		printf ("Send key2 to LU failed (expected)\n");
		return -1;
	}

	if (ai.type == DVD_AUTH_ESTABLISHED)
		printf ("DVD is authenticated\n");
	else if (ai.type == DVD_AUTH_FAILURE)
		printf ("DVD authentication failed\n");

	memcpy(Challenge, Key1.b, 5);
	memcpy(Challenge+5, Key2.b, 5);
	CryptBusKey(varient, Challenge, &KeyCheck);

#ifdef DEBUG
	printf("Received Session Key:\t");
	for (i= 0; i< 5; i++) {
		printf ("%02X ", KeyCheck.b[i]);
	}
	printf("\n");
#endif

	if (CSSGetASF (fd) < 0)
		return -1;

	_get_title_key (fd, agid, lba, KeyCheck.b, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}
