/*  Livid
 *
 * Configuration dialog
 *
 * Copyright (C) 1999  Thomas Mirlacher, Guenter Bartsch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define __USE_LARGEFILE64

#include <stdio.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include <nav/DVDPlayerAPI.h>

#include "config_devices.h"

#include "main.h"

extern struct player_global_struct player_global;
GtkWidget *conf_win;


/**
 *
 */

void close_config ()
{
  gtk_widget_hide (conf_win);
  player_global.bConfShowing = FALSE;

}

void config_ok (GtkButton *button,
		gpointer user_data) 
{
  config_device_apply ();
  save_config ();
  close_config ();
}

void config_apply (GtkButton *button,
		   gpointer user_data) 
{
  config_device_apply ();
}

void config_cancel (GtkButton *button,
		    gpointer user_data) 
{
  read_config ();
  close_config ();
}

/**
 *
 */

GtkWidget* setup_config (void)
{

  GtkWidget *button;
  GtkWidget *vbox, *hbox;
  GtkWidget *notebook;
  GtkWidget *label, *label2;
  

  conf_win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT(conf_win), "delete_event",
		      GTK_SIGNAL_FUNC (close_config), NULL);

  gtk_window_set_title(GTK_WINDOW(conf_win), "Configuration");
  

  gtk_container_set_border_width (GTK_CONTAINER (conf_win), 5);

  // vbox : notebook top, buttons bottom

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (conf_win), vbox);
  
  /* Create a new notebook, place the position of the tabs */
  notebook = gtk_notebook_new ();
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
  gtk_container_set_border_width (GTK_CONTAINER (notebook), 5);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);
  gtk_widget_show(notebook);


  /*
   * Device-page
   */
  
  label = gtk_label_new ("Devices");
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), create_devices_page (), label);

  /*
   * Video-Page
   */
  
  label2 = gtk_label_new ("not implemented");
  gtk_widget_show (label2);
    
  label = gtk_label_new ("Video");
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), label2, label);

  /*
   * Audio-Page
   */
  
  label2 = gtk_label_new ("not implemented");
  gtk_widget_show (label2);
    
  label = gtk_label_new ("Audio");
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), label2, label);

  /*
   * buttons
   */

  hbox = gtk_hbox_new (TRUE, 5);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show(hbox);

  button = gtk_button_new_with_label ("OK");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) config_ok,
			     GTK_OBJECT (notebook));
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  button = gtk_button_new_with_label ("Apply");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) config_apply,
			     GTK_OBJECT (notebook));
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  button = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) config_cancel,
			     GTK_OBJECT (notebook));
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  gtk_widget_show(vbox);
  
  return conf_win;
}







