#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include <pthread.h>

#include <udf/dvd_udf.h>
#include <css/css.h>
#include <ifo/libifo/ifo.h>

#include "nav.h"
#include <plugin/plugin.h>

#define DEVICE_NAME_LEN 8
#define DEVICE_VCD "/dev/vcd"
#define DEVICE_DVD "/dev/dvd"
#define DEVICE_DVB "/dev/dvb"

typedef struct {
	char *root;		// rootpath or file

	struct {
		int audio;
		int subpic;

		int title;
		int chapter;

		int status;
	} current;

	struct {
		pthread_mutex_t mutex;
		pthread_cond_t playing;
	} ctrl;

	struct {
		void (*chapter) (void *data, int chapter);
		void *chapter_data;
	} callb;        // Callbacks

	pthread_t play_loop;

	int fd;

	plugin_t plugin;

	buf_t buf;
} priv_t;

static priv_t foo;
priv_t *priv = &foo;


/**
 * open input device/file and associated plugins
 */

int navOpenInput (char *path)
{
	struct stat info;
	void *handle;
	int (*nav) (plugin_t *nav, char *path);

// just to be sure to have old stuff closed, before heading for new things 
	navCloseInput ();

	if (stat (path, &info) == -1)
		return DVD_PLAYER_NOT_FOUND;

fprintf (stderr, "DEBUG: navOpenInput: path (%s) found\n", path);

	priv->current.status = STATUS_STOP;
	priv->root = path;

	if (S_ISCHR(info.st_mode) || S_ISBLK (info.st_mode)) { // device
		if (!strncmp (path, DEVICE_DVD, DEVICE_NAME_LEN)) {
			if (pluginOpen (&priv->plugin, PLUGIN_ID_NAV, DL_NAV_DVD, NULL) < 0)
				return -1;
fprintf (stderr, "I'M OK ... go ahead\n");
			return PLUGIN_OPS_NAV(&priv->plugin)->open (&priv->plugin, path);
		} else if (!strncmp (path, DEVICE_VCD, DEVICE_NAME_LEN)) {
			if (pluginOpen (&priv->plugin, PLUGIN_ID_NAV, DL_NAV_VCD, NULL) < 0)
				return -1;
			return PLUGIN_OPS_NAV(&priv->plugin)->open (&priv->plugin, path);
		} else if (!strncmp (path, DEVICE_DVB, DEVICE_NAME_LEN)) {
			fprintf (stderr, "found DVB device\n");
			return -1;
		}
	} else if (S_ISREG (info.st_mode)) {        // file
		if (pluginOpen (&priv->plugin, PLUGIN_ID_NAV, DL_NAV_FILE, NULL) < 0)
			return -1;
		return PLUGIN_OPS_NAV(&priv->plugin)->open (&priv->plugin, path);
	} else {
		fprintf (stderr, "UNKNOWN\n");
	}

#if 0
	   if (!plugin->codec) {
                fprintf (stderr, "no codec regstered!");
        } else
                plugin->decaps->write = plugin->codec->read;
#endif
        return DVD_PLAYER_NOT_FOUND;
}


/**
 * close current input device/file and associated plugins
 */

int navCloseInput (void)
{
	// stop player if still running
	navSetStatus (STATUS_STOP);

	// clean up plugins
	pluginClose (&priv->plugin, PLUGIN_ID_NAV);
	pluginClose (&priv->plugin, PLUGIN_ID_DECAPS);
	pluginClose (&priv->plugin, PLUGIN_ID_INPUT);

	return DVD_PLAYER_OK;
}


/**
 * open output device (and load relevant plugins)
 */

int navOpenOutput (char *plugin_path, char *output_path)
{
	void *handle;
	plugin_codec_t *(*codec) (void);

	if (!(handle = _pluginOpen (&priv->plugin, PLUGIN_ID_CODEC, plugin_path)))
		return -1;

	codec = dlsym (handle, "codec_init");
	PLUGIN_OPS_CODEC(&priv->plugin) = (*codec) ();

	if (!PLUGIN_OPS_DECAPS(&priv->plugin))
		fprintf (stderr, "no decapsulation registered\n");
	else
		PLUGIN_OPS_DECAPS(&priv->plugin)->write = PLUGIN_OPS_CODEC(&priv->plugin)->read;

	return PLUGIN_OPS_CODEC(&priv->plugin)->init (&priv->buf, output_path); 
}


/**
 * clean all we've messed around during outputting). ok - frankly speaking,
 * this still needs a little bit of work.
 */

int navCloseOutput (void)
{
	pluginClose (&priv->plugin, PLUGIN_ID_CODEC);
	return 0;
}


/**
 *
 */

nav_tree_t *navGetInfo (void)
{
	return PLUGIN_OPS_NAV(&priv->plugin)->get_info ();
}


/**
 * Main Loop.
 */

void *navLoop (void *foo)
{
fprintf (stderr, "DEBUG: navLoop: decoder loop started\n");
	PLUGIN_OPS_CODEC(&priv->plugin)->buf_init (&priv->buf);
	PLUGIN_OPS_CODEC(&priv->plugin)->set_status (STATUS_PLAY);

	while ((priv->current.status != STATUS_STOP) &&
	       (priv->current.status != STATUS_FINISH)) {
		if (priv->current.status == STATUS_PAUSE) {
			PLUGIN_OPS_CODEC(&priv->plugin)->set_status (STATUS_PAUSE);
			pthread_cond_wait (&priv->ctrl.playing, &priv->ctrl.mutex);
			PLUGIN_OPS_CODEC(&priv->plugin)->set_status (STATUS_PLAY);
		}
		if (PLUGIN_OPS_DECAPS(&priv->plugin)->decaps (&priv->buf)) {
			priv->current.status = STATUS_FINISH;
			break;
		}
	}

fprintf (stderr, "DEBUG: navLoop: decoder loop stopped\n");
	PLUGIN_OPS_CODEC(&priv->plugin)->set_status (STATUS_STOP);
fprintf (stderr, "DEBUG: navLoop: decoder stopped\n");
	PLUGIN_OPS_CODEC(&priv->plugin)->buf_exit (&priv->buf);
fprintf (stderr, "DEBUG: navLoop: buf exit\n");

// call callback here ... and inform listeners, we've changed our state 

	if (priv->callb.chapter)
		priv->callb.chapter (priv->callb.chapter_data, 0);

	pthread_exit (NULL);
}


/**
 * Register callback
 */

int navRegisterCb (u_int id, void (*cb) (void *, int ), void *data)
{
	priv->callb.chapter = cb;
	priv->callb.chapter_data = data;

	if (PLUGIN_OPS_NAV(&priv->plugin)->register_callb) {
		PLUGIN_OPS_NAV (&priv->plugin)->register_callb (id, cb, data);
		return 0;
	}

	return -1;
}


/**
 * Unregister callback
 */

int navUnregisterCb (u_int id)
{
	if (PLUGIN_OPS_NAV(&priv->plugin)->unregister_callb) {
		PLUGIN_OPS_NAV (&priv->plugin)->unregister_callb (id);
		return 0;
	}

	return -1;
}


/**
 * Play the specified title of the DVD
 * Title:  1..99
 */

int TitlePlay (int title)
{
	return PartPlayAutoStop (title, 0, 0);
} 


/**
 * Play the specified title of the DVD, from the specified chapter on
 * Title:  1..99
 * Part:   1..999
 */

int PartPlay (int title, int chapter)
{
	return PartPlayAutoStop (title, chapter, 0);
}


/**
 *
 */

int PartPlayAutoStop (int title, int chapter, int parts2play)
{
        int ret = PLUGIN_OPS_NAV(&priv->plugin)->play (title, chapter, parts2play);

fprintf (stderr, "DEBUG: PLAY\n");

        if (priv->current.status == STATUS_STOP) {
                pthread_mutex_init(&priv->ctrl.mutex, NULL);
                pthread_cond_init (&priv->ctrl.playing, NULL);

                priv->current.status = STATUS_PLAY;

                pthread_create (&priv->play_loop, NULL, navLoop, NULL);
        } else {
                priv->current.status = STATUS_PLAY;
                pthread_cond_signal (&priv->ctrl.playing);
        }

	return ret;
}


/**
 * set staus of navigator
 */

int navSetStatus (u_int status)
{
//fprintf (stderr, "DEBUG: navSetStatus %d\n", status);

	switch (status) {
		case STATUS_PAUSE_ON:
			if (priv->current.status != STATUS_PLAY)
				return DVD_PLAYER_WRONG_DOMAIN;

			priv->current.status = STATUS_PAUSE;
			break;
		case STATUS_PAUSE_OFF:
			if (priv->current.status != STATUS_PAUSE)
				return DVD_PLAYER_WRONG_DOMAIN;

			priv->current.status = STATUS_PLAY;
			pthread_cond_signal (&priv->ctrl.playing);
			break;
		case STATUS_STOP:
			if (priv->current.status == STATUS_STOP)
				return DVD_PLAYER_WRONG_DOMAIN;

			priv->current.status = STATUS_STOP;
			pthread_cancel (priv->play_loop);
//pthread_cond_signal (&priv->ctrl.playing);

			pthread_join (priv->play_loop, NULL);
			pthread_mutex_destroy (&priv->play_loop);

			// close open descriptors here
			break;
		default:
			return DVD_PLAYER_NOT_AVAILABLE;
	} 

	return DVD_PLAYER_OK;
}


/**
 * get internal information of navigator status
 */

int navGetStatus (void)
{
	return priv->current.status;
}


/**
 *
 */

int GetRoot (char *Root, int BufSize, int *ActualSize)
{
	strncpy (Root,priv->root,BufSize);
	*ActualSize=strlen (priv->root);

	return DVD_PLAYER_OK;
}


/**
 *
 */

int TimeSearch (u_long TimeCode)
{
	int hour,min,sec,frame,rate;
	TimeCode2Time(TimeCode,hour,min,sec,frame,rate);

	return DVD_PLAYER_OK;
}


/**
 *
 */

int GetTotalTitleTime (u_long *TotalTime)
{
	int tot=0;		//=_player_total_time();
	if (tot == -1)
		return DVD_PLAYER_ERROR;

	*TotalTime = Time2TimeCode(tot/3600,tot%3600/60,tot%60,0,0);
	return DVD_PLAYER_OK;
}



/**
 *
 */

int GetCurrentLocation (int *TitleNum, int *PartNum, u_long *TimeCode)
{
	int time=0;	//=_player_current_time();

	if (time == -1)
		return DVD_PLAYER_WRONG_DOMAIN;

	*TitleNum = priv->current.title;
	*PartNum  = priv->current.chapter;
	*TimeCode = Time2TimeCode(time/3600,time%3600/60,time%60,0,0);

	return DVD_PLAYER_OK;
}


/**
 *
 */

int VideoPresentationModeChange (int InitialDisplay, int CurrentDisplay)
{
	plugin_codec_attr_video_t attr = {-1, -1, CurrentDisplay, InitialDisplay, -1};

	if (InitialDisplay != 0 && InitialDisplay != 3 &&
		CurrentDisplay >= 0 && CurrentDisplay <= 2)
		return DVD_PLAYER_INVALID;

	if (PLUGIN_OPS_CODEC(&priv->plugin)->set_attributes (BUF_VIDEO, &attr))
		return DVD_PLAYER_ERROR;

	return DVD_PLAYER_OK;
}


/**
 *
 */

int PartSearch (int Part)
{
	if (Part < 1 || Part > 999)
		return DVD_PLAYER_INVALID;

	return DVD_PLAYER_OK;
}


/**
 * Select specified audio stream
 * Audio: 0..7
 */

int AudioStreamChange (int Audio)
{
	plugin_codec_attr_audio_t attr = {-1, -1, -1, Audio, -1};

	if (Audio < 0 || Audio > 7)
		return DVD_PLAYER_INVALID;

	priv->current.audio = Audio;

	if (PLUGIN_OPS_DECAPS(&priv->plugin)->set_filter (BUF_AUDIO, Audio))
		return DVD_PLAYER_ERROR;

	if (PLUGIN_OPS_CODEC(&priv->plugin)->set_attributes  (BUF_AUDIO, &attr))
		return DVD_PLAYER_ERROR;

	return 0;
}



/**
 * Select specified sub picture stream and set if sub picture
 *	should be displayed or not
 * Subpicture: 0..31 or 63 (dummy stream)
 * Display: 0 or 1
 */

int SubpictureStreamChange (int Subpicture, int Disabled)
{
	plugin_codec_attr_spu_t spu;
	spu.clut = NULL;

	if (Disabled == 1 || Subpicture == 63)
		spu.num = -1;
	else if (Subpicture >= 0 && Subpicture <= 31)
		spu.num = Subpicture;
	else
		return DVD_PLAYER_INVALID;

	if (PLUGIN_OPS_DECAPS(&priv->plugin)->set_filter (BUF_SUBPIC, spu.num))
		return DVD_PLAYER_ERROR;

	if (PLUGIN_OPS_CODEC(&priv->plugin)->set_attributes (BUF_SUBPIC, &spu))
		return DVD_PLAYER_ERROR;

	priv->current.subpic = spu.num;

	return DVD_PLAYER_OK;
}


/**
 *
 */

int GetAudioLanguage(int Stream, char *LanguageCode)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}


/**
 *
 */

int GetCurrentAudio(int *StreamsAvailable, int *CurrentStream)
{
	*StreamsAvailable = 8;
	*CurrentStream = priv->current.audio;

	return DVD_PLAYER_OK;
}


/**
 *
 */

int GetCurrentAudioAttributes(u_char *ATR)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}


/**
 *
 */

int GetCurrentSubpicture(int *StreamsAvailable, int *CurrentStream, int *IsDisabled)
{
	*StreamsAvailable=10;
	if (priv->current.subpic == -1) {
		*CurrentStream=63;
		*IsDisabled=1;
	} else {
		*CurrentStream=priv->current.subpic;
		*IsDisabled=0;
	}

	return DVD_PLAYER_OK;
}


/**
 *
 */

int GetCurrentSubpictureAttributes(u_char *ATR)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}


/**
 *
 */

int GetSubpictureLanguage(int Stream, char *LanguageCode)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}


/**
 *
 */

int AngleChange(int Angle)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}


/**
 *
 */

int GetCurrentAngle(int *AnglesAvailable, int *CurrentAngle)
{
	return DVD_PLAYER_NOT_AVAILABLE;
}
