#include <stdio.h>
#include <dlfcn.h>
#include <string.h>
#include <sys/types.h>

#include "nav.h"
#include <plugin/plugin.h>


/**
 * Open and initialize plugin
 */

int pluginOpen (plugin_t *plugin, u_int plugin_id, char *plugin_name, void *data)
{
	void *(*init)(void *);

	if (plugin_id >= MAX_PLUGIN_ID) {
		fprintf (stderr, "unknown plugin type\n");
		return -1;
	}

	if (!(plugin->handle[plugin_id] = dlopen (plugin_name, RTLD_LAZY))) {
		fprintf (stderr, "error in opening plugin: %s\n", dlerror ());
		fprintf (stderr, "do you have set your LD_LIBRARY_PATH correctly?\n");

		return -1;
	}

	init = dlsym (plugin->handle[plugin_id], "plugin_init");
	if (!(plugin->ops[plugin_id] = (*init) (data)))
		return -1;

	return 0;
}


/**
 * Open plugin
 */

void *_pluginOpen (plugin_t *plugin, u_int plugin_id, char *plugin_name)
{
	if (plugin_id >= MAX_PLUGIN_ID) {
		fprintf (stderr, "unknown plugin type\n");
		return NULL;
	}

	if (!(plugin->handle[plugin_id] = dlopen (plugin_name, RTLD_LAZY))) {
		fprintf (stderr, "error in opening plugin: %s\n", dlerror ());
		fprintf (stderr, "do you have set your LD_LIBRARY_PATH correctly?\n");

		return NULL;
	}

	return plugin->handle[plugin_id];
}


/**
 * Close plugin and clean up plugin structure 
 */

void *pluginClose (plugin_t *plugin, u_int plugin_id)
{
	if (plugin_id >= MAX_PLUGIN_ID) {
		fprintf (stderr, "unknown plugin type\n");
		return NULL;
	}

//DENT call plugin->*->close() here ...
	if (plugin->handle[plugin_id])
		dlclose (plugin->handle[plugin_id]);
	plugin->ops[plugin_id] = NULL;

	return NULL;
}


/**
 *
 */

// DENT: should work as follows (in the future)
//	1) compare fileextension
//		if extension matches, check if magic also matches
//	2) if no filename matches, check all extensions
//
//	OR ... just use it like now - just check magic codes
//	i think i'll implment it like now, but save all magic codes into a list
//	so i don't have to dlopen the plugin first.
 
static int _pluginGetDecapsMagic (plugin_t *plugin, char *plugin_name, u_char *buf)
{
	plugin_decaps_magic_t *magic;
	int i;

	pluginOpen (plugin, PLUGIN_ID_DECAPS, plugin_name, NULL);
	magic = PLUGIN_OPS_DECAPS(plugin)->magic ();

	if (magic) {
		fprintf (stderr, "%s: file: %s\n", plugin_name, magic->file_ext);
		fprintf (stderr, "%s: len: %d\n", plugin_name, magic->magic_len);

		fprintf (stderr, "%s: magic: ", plugin_name);
		for (i=0; i<magic->magic_len; i++)
			fprintf (stderr, "0x%02x ", magic->magic[i]);
		fprintf (stderr, "\n\n");

		if (!(strncmp (buf, magic->magic, magic->magic_len))) {
			fprintf (stderr, "%s matches -> found the right decapsulation for out file\n", plugin_name);
			return 0;
		}

	} else
		fprintf (stderr, "%s: NO MAGIC!\n\n", plugin_name);

	pluginClose (plugin, PLUGIN_ID_DECAPS);

	return -1;
}


/**
 * DENT: i know this is an ugly contruct right now, but:
 *	1) if we've an install routine, plugins will be looked up
 *		in the directory
 *	2) should use PLUGIN_OPS_INPUT 
 */

int pluginGetDecapsMagic (plugin_t *plugin, char *filename)
{
	u_char buf[20];
	int i;

	PLUGIN_OPS_INPUT(plugin)->seek (0);
	PLUGIN_OPS_INPUT(plugin)->read (buf, 20);

	for (i=0; i<20; i++)
		fprintf (stderr, "0x%02x ", buf[i]);

	if (!_pluginGetDecapsMagic (plugin, DL_DECAPS_PS, buf))
		return 0;
	if (!_pluginGetDecapsMagic (plugin, DL_DECAPS_TS, buf))
		return 0;
	if (!_pluginGetDecapsMagic (plugin, DL_DECAPS_VCD, buf))
		return 0;

	return -1;
} 

